import copy
import functools
import logging

from collections import defaultdict
from enum import Enum
from typing import Union, List, Callable, Optional, Dict, Any, Tuple, Iterable

import numpy as np

import pyarrow as pa
import pyarrow.fs
from pyarrow.fs import S3FileSystem

from ray.data import (
    Datasource,
    ReadTask,
)
from ray.data.block import BlockMetadata, Block, BlockAccessor
from ray.data.datasource import (
    FastFileMetadataProvider,
    ParquetMetadataProvider,
)

from deltacat.constants import METAFILE_FORMAT_MSGPACK
from deltacat.aws.s3u import (
    S3Url,
    parse_s3_url,
)
from deltacat.types.media import (
    ContentType,
)
from deltacat.storage import (
    Manifest,
    ManifestEntryList,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.url import DeltaCatUrl, DeltaCatUrlReader
from deltacat.storage import (
    Metafile,
    ListResult,
)
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

METAFILE_DATA_COLUMN_NAME = "deltacat_metafile_data"
METAFILE_TYPE_COLUMN_NAME = "deltacat_metafile_type"


class DeltacatReadType(str, Enum):
    METADATA = "metadata"  # get only a single metafile
    METADATA_LIST = "metadata_list"  # list top-level metafiles
    METADATA_LIST_RECURSIVE = "metadata_recursive"  # list all metafiles
    DATA = "data"  # read all data files


class CachedFileMetadataProvider(
    FastFileMetadataProvider,
    ParquetMetadataProvider,
):
    def __init__(self, meta_cache: Dict[str, BlockMetadata]):
        self._meta_cache = meta_cache

    def get_meta_cache(self) -> Dict[str, BlockMetadata]:
        return self._meta_cache

    def _get_block_metadata(
        self,
        paths: List[str],
        schema: Optional[Union[type, pa.Schema]],
        **kwargs,
    ) -> BlockMetadata:
        agg_block_metadata = BlockMetadata(
            num_rows=0,
            size_bytes=0,
            schema=schema,
            input_files=[],
            exec_stats=None,
        )
        for path in paths:
            block_metadata = self._meta_cache.get(path)
            if block_metadata is None:
                raise ValueError(f"Block metadata not found for path: {path}")
            if block_metadata.num_rows is None:
                agg_block_metadata.num_rows = None
            elif agg_block_metadata.num_rows is not None:
                agg_block_metadata.num_rows += block_metadata.num_rows
            if block_metadata.size_bytes is None:
                agg_block_metadata.size_bytes = None
            elif agg_block_metadata.size_bytes is not None:
                agg_block_metadata.size_bytes += block_metadata.size_bytes
            agg_block_metadata.input_files.append(path)
        return agg_block_metadata


class PathType(str, Enum):
    MANIFEST = "manifest"
    FILES_AND_FOLDERS = "files_and_folders"


class DelimitedTextReaderConfig:
    def __init__(
        self,
        csv: bool = False,
        header: bool = False,
        delimiter: Optional[str] = None,
        bzip2: bool = False,
        gzip: bool = False,
        zstd: bool = False,
        add_quotes: Optional[bool] = None,
        null_as: str = "",
        escape: bool = False,
        fixed_width: bool = False,
    ):
        self.header = header
        self.delimiter = delimiter if delimiter else "," if csv else "|"
        self.bzip2 = bzip2
        self.gzip = gzip
        self.zstd = zstd
        self.add_quotes = add_quotes if add_quotes else True if csv else False
        self.null_as = null_as
        self.escape = escape
        self.fixed_width = fixed_width

    def _get_arrow_compression_codec_name(self) -> str:
        arrow_compression_codec_name = None
        codecs_enabled = {
            "bz2": self.bzip2,
            "gzip": self.gzip,
            "zstd": self.zstd,
        }
        for encoding, flag in codecs_enabled.items():
            if arrow_compression_codec_name and flag:
                raise ValueError(
                    f"Multiple delimited text compression types specified "
                    f"({codecs_enabled}). Please ensure that only one "
                    f"compression type is set and try again."
                )
            if flag:
                arrow_compression_codec_name = encoding
        return arrow_compression_codec_name

    def to_arrow_reader_kwargs(
        self, include_columns: Optional[List[str]], schema: Optional[pa.Schema]
    ) -> Dict[str, Any]:
        from pyarrow import csv

        if self.fixed_width:
            raise NotImplementedError(
                "Delimited text files configured with FIXEDWIDTH are not "
                "currently supported."
            )
        open_stream_args = {}
        arrow_compression_codec_name = self._get_arrow_compression_codec_name()
        if arrow_compression_codec_name:
            open_stream_args["compression"] = arrow_compression_codec_name
        column_names = None
        if schema:
            column_names = schema.names
        autogen_column_names = False if self.header or column_names else True
        read_options = csv.ReadOptions(
            use_threads=False,
            column_names=column_names,
            autogenerate_column_names=autogen_column_names,
        )
        parse_options = csv.ParseOptions(
            delimiter=self.delimiter,
            quote_char='"' if self.add_quotes else False,
            escape_char="\\" if self.escape else False,
            double_quote=False if self.escape else True,
        )
        convert_options = csv.ConvertOptions(
            column_types=schema,
            null_values=[self.null_as] if self.null_as is not None else [],
            true_values=["t"],
            false_values=["f"],
            strings_can_be_null=True if self.null_as is not None else False,
            quoted_strings_can_be_null=True if self.null_as else False,
            include_columns=include_columns,
        )
        return {
            "open_stream_args": open_stream_args,
            "read_options": read_options,
            "parse_options": parse_options,
            "convert_options": convert_options,
        }


def normalize_s3_paths_for_filesystem(
    paths: Union[str, List[str]],
    filesystem: pyarrow.fs.FileSystem,
) -> Tuple[List[str], List[S3Url]]:
    urls = []
    if isinstance(paths, str):
        paths = [paths]
    if isinstance(filesystem, S3FileSystem):
        urls = [parse_s3_url(url) for url in paths]
        # pyarrow.fs.FileSystem paths should not start with "s3://"
        # pyarrow.fs.FileSystem paths should not end with "/"
        paths = [f"{u.bucket}/{u.key}".rstrip("/") for u in urls]
    return paths, urls


def _infer_content_types_from_paths(
    paths: List[str],
    content_type_provider: Callable[[str], ContentType],
) -> Dict[ContentType, List[str]]:
    content_type_to_paths = defaultdict(list)
    for path in paths:
        if not path.endswith("/"):
            content_type_to_paths[content_type_provider(path)].append(path)
    return content_type_to_paths


def _expand_manifest_paths_by_content_type(
    manifest: Manifest,
) -> Tuple[Dict[ContentType, List[str]], CachedFileMetadataProvider]:
    content_type_to_paths = {}
    meta_provider = CachedFileMetadataProvider({})
    if not manifest.entries:
        logger.warning(f"No entries to read in DeltaCAT Manifest: {manifest}")
    else:
        content_type_to_paths, meta_provider = _read_manifest_entry_paths(
            manifest.entries,
            manifest.meta.content_type if manifest.meta else None,
        )
    # TODO(pdames): infer the schema from a manifest if available?
    # if not schema and ContentType.PARQUET not in content_type_to_paths:
    #     schema = _infer_schema_from_manifest(manifest)
    return content_type_to_paths, meta_provider


def _read_manifest_entry_paths(
    entries: ManifestEntryList,
    manifest_content_type: Optional[str],
) -> Tuple[Dict[ContentType, List[str]], CachedFileMetadataProvider]:
    # support manifests with heterogenous content types
    content_type_to_paths = defaultdict(list)
    meta_cache: Dict[str, BlockMetadata] = {}
    for e in entries:
        url = e.url if e.url else e.uri
        # get manifest entry content type or fall back to manifest content type
        content_type = e.meta.content_type or manifest_content_type
        if content_type:
            content_type_to_paths[ContentType(content_type)] = url
        else:
            # TODO(pdames): fall back to content type inference by file extension
            raise ValueError(
                f"Manifest entry missing content type: {e}. "
                f"Please specify a content type for each manifest entry."
            )
        meta_cache[url] = BlockMetadata(
            num_rows=e.meta.record_count,
            size_bytes=e.meta.content_length,
            schema=None,
            input_files=[],
            exec_stats=None,
        )
    return content_type_to_paths, CachedFileMetadataProvider(meta_cache)


def _get_metafile_read_task(
    metafile: Metafile,
) -> Iterable[Block]:
    pyarrow_table_dict = {
        METAFILE_DATA_COLUMN_NAME: [metafile.serialize(METAFILE_FORMAT_MSGPACK)],
        METAFILE_TYPE_COLUMN_NAME: [Metafile.get_type_name(metafile)],
    }
    yield BlockAccessor.batch_to_arrow_block(pyarrow_table_dict)


def _get_metafile_lister_read_task(
    lister: Callable[[Any], ListResult[Metafile]],
    all_lister_kwargs: List[Dict[str, Any]],
) -> Iterable[Block]:
    metafiles = []
    for lister_kwargs in all_lister_kwargs:
        metafile_list_result = lister(**lister_kwargs)
        # TODO(pdames): switch to paginated read
        metafiles.append(metafile_list_result.all_items())
    pyarrow_table_dict = {
        METAFILE_DATA_COLUMN_NAME: [
            meta.serialize(METAFILE_FORMAT_MSGPACK)
            for metasublist in metafiles
            for meta in metasublist
        ],
        METAFILE_TYPE_COLUMN_NAME: [
            Metafile.get_class(meta).__name__
            for metasublist in metafiles
            for meta in metasublist
        ],
    }
    yield BlockAccessor.batch_to_arrow_block(pyarrow_table_dict)


class DeltaCatDatasource(Datasource):
    """Datasource for reading registered DeltaCAT catalog objects."""

    def __init__(
        self,
        url: DeltaCatUrl,
        deltacat_read_type: DeltacatReadType = DeltacatReadType.DATA,
        timestamp_as_of: Optional[int] = None,
        merge_on_read: Optional[bool] = False,
        read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ):
        self._url = url
        self._reader = DeltaCatUrlReader(url)
        self._deltacat_read_type = deltacat_read_type
        self._timestamp_as_of = timestamp_as_of
        self._merge_on_read = merge_on_read
        self._filesystem = url.catalog.filesystem
        self._read_kwargs_provider = read_kwargs_provider

    def estimate_inmemory_data_size(self) -> Optional[int]:
        """Return an estimate of the in-memory data size, or None if unknown.

        Note that the in-memory data size may be larger than the on-disk data size.
        """
        return None

    def get_read_tasks(self, parallelism: int) -> List[ReadTask]:
        """Execute the read and return read tasks.

        Args:
            parallelism: The requested read parallelism. The number of read
                tasks should equal to this value if possible.

        Returns:
            A list of read tasks that can be executed to read blocks from the
            datasource in parallel.
        """
        kwargs = self._read_kwargs_provider(self._url.datastore_type, {})
        if self._deltacat_read_type == DeltacatReadType.METADATA:
            # do a shallow read of the top-level DeltaCAT metadata
            empty_block_metadata = BlockMetadata(
                num_rows=None,
                size_bytes=None,
                schema=None,
                input_files=None,
                exec_stats=None,
            )
            metafile = self._reader.read(**kwargs)
            read_tasks = [
                ReadTask(
                    lambda: _get_metafile_read_task(metafile),
                    empty_block_metadata,
                )
            ]
        elif self._deltacat_read_type == DeltacatReadType.METADATA_LIST:
            # do a shallow read of the top-level DeltaCAT metadata
            listers = copy.deepcopy(self._reader.listers)
            listers = [listers[0]]
            read_tasks = self._list_all_metafiles_read_tasks(
                parallelism=parallelism,
                listers=listers,
                **kwargs,
            )
        elif self._deltacat_read_type == DeltacatReadType.METADATA_LIST_RECURSIVE:
            read_tasks = self._list_all_metafiles_read_tasks(
                parallelism=parallelism,
                listers=copy.deepcopy(self._reader.listers),
                **kwargs,
            )

        elif self._deltacat_read_type == DeltacatReadType.DATA:
            # do a deep read across all in-scope Delta manifest file paths
            # recursive is implicitly true for deep data reads
            # TODO(pdames): For data reads targeting DeltaCAT catalogs, run a
            #  recursive distributed metadata read first, then a data read
            #  second.
            raise NotImplementedError()
            """
            list_results = self._list_all_metafiles(**kwargs)
            deltas: List[Delta] = list_results[len(list_results) - 1]
            read_tasks = []
            for delta in deltas:
                read_tasks.append(
                    self._get_delta_manifest_read_tasks(
                        delta.manifest,
                        parallelism,
                    ),
                )
            """
        else:
            raise NotImplementedError(
                f"Unsupported DeltaCAT read type: {self._deltacat_read_type}"
            )

        return read_tasks

    def _list_all_metafiles_read_tasks(
        self,
        parallelism: int,
        listers: List[Callable[[Any], ListResult[Metafile]]],
        **kwargs,
    ) -> List[ReadTask]:
        list_results: List[ListResult[Metafile]] = []
        # the first lister doesn't have any missing keyword args
        (
            first_lister,
            first_kwarg_name,
            first_kwarg_val_resolver_fn,
        ) = listers.pop(0)
        if listers:
            metafile_list_result = first_lister(**kwargs)
            list_results.append(metafile_list_result)
            (
                last_lister,
                last_kwarg_name,
                last_kwarg_val_resolver_fn,
            ) = listers.pop()
        else:
            metafile_list_result = None
            (
                last_lister,
                last_kwarg_name,
                last_kwarg_val_resolver_fn,
            ) = (first_lister, first_kwarg_name, first_kwarg_val_resolver_fn)
        for lister, kwarg_name, kwarg_val_resolver_fn in listers:
            # each subsequent lister needs to inject missing keyword args from the parent metafile
            for metafile in metafile_list_result.all_items():
                kwargs_update = (
                    {kwarg_name: kwarg_val_resolver_fn(metafile)}
                    if kwarg_name and kwarg_val_resolver_fn
                    else {}
                )
                lister_kwargs = {
                    **kwargs,
                    **kwargs_update,
                }
                metafile_list_result = lister(**lister_kwargs)
                list_results.append(metafile_list_result)
        empty_block_metadata = BlockMetadata(
            num_rows=None,
            size_bytes=None,
            schema=None,
            input_files=None,
            exec_stats=None,
        )
        if metafile_list_result:
            # use a single read task to materialize all prior metafiles read
            # as an arrow table block
            # (very lightweight, so not counted against target parallelism)
            read_tasks = [
                ReadTask(
                    read_fn=functools.partial(
                        _get_metafile_lister_read_task,
                        lister=lambda all_list_results: ListResult.of(
                            [
                                item
                                for list_result in all_list_results
                                for item in list_result.all_items()
                            ]
                        ),
                        all_lister_kwargs=[{"all_list_results": list_results}],
                    ),
                    metadata=empty_block_metadata,
                )
            ]
            # parallelize the listing of all metafile leaf nodes
            split_metafiles = np.array_split(
                metafile_list_result.all_items(),
                parallelism,
            )
            for metafiles in split_metafiles:
                all_lister_kwargs = []
                for metafile in metafiles:
                    kwargs_update = (
                        {last_kwarg_name: last_kwarg_val_resolver_fn(metafile)}
                        if last_kwarg_name and last_kwarg_val_resolver_fn
                        else {}
                    )
                    lister_kwargs = {
                        **kwargs,
                        **kwargs_update,
                    }
                    all_lister_kwargs.append(lister_kwargs)
                read_tasks.append(
                    ReadTask(
                        read_fn=functools.partial(
                            _get_metafile_lister_read_task,
                            lister=last_lister,
                            all_lister_kwargs=all_lister_kwargs,
                        ),
                        metadata=empty_block_metadata,
                    )
                )
        else:
            # first lister is also the last lister (i.e., shallow listing)
            read_tasks = [
                ReadTask(
                    read_fn=functools.partial(
                        _get_metafile_lister_read_task,
                        lister=last_lister,
                        all_lister_kwargs=[kwargs],
                    ),
                    metadata=empty_block_metadata,
                )
            ]
        return read_tasks

    """

    def _get_delta_manifest_read_tasks(
        self,
        delta_manifest: Manifest,
        parallelism: int,
    ) -> List[ReadTask]:
        # find all files in the Delta manifest
        content_type_to_paths, meta_provider = _expand_manifest_paths_by_content_type(
            delta_manifest,
            self._filesystem,
        )
        num_content_types = len(content_type_to_paths)
        if num_content_types > 1 and not schema:
            # infer schema from a single parquet file
            # TODO (pdames): read verbose manifest schema if available, and infer
            #  schema from a sample parquet dataset if not
            path = content_type_to_paths[ContentType.PARQUET][0]
            with resolved_fs.open_input_file(path, **open_stream_args) as f:
                schema = pq.read_schema(f)
        content_type_to_reader = {
            ContentType.PARQUET: ParquetDatasource(),
            ContentType.CSV: CSVDatasource(),
        }
        all_read_tasks = []
        for content_type, paths in content_type_to_paths.items():
            reader = content_type_to_reader.get(content_type)
            assert reader, f"No datasource found for: {content_type}"
            prepare_read_kwargs = {
                "parallelism": parallelism,
                "paths": paths,
                "filesystem": self._filesystem,
                "schema": schema,
                "meta_provider": meta_provider,
            }
            if content_type == ContentType.PARQUET:
                if columns:
                    prepare_read_kwargs["columns"] = columns
            elif content_type in DELIMITED_TEXT_CONTENT_TYPES:
                prepare_read_kwargs.update(
                    csv_reader_config.to_arrow_reader_kwargs(columns, schema)
                )
            else:
                raise NotImplementedError(f"Unsupported content type: {content_type}")
            # merge any provided reader kwargs for this content type with those
            # inferred from CSV Reader Config
            if read_kwargs_provider:
                prepare_read_kwargs = read_kwargs_provider(
                    content_type,
                    prepare_read_kwargs,
                )
            # explicitly specified `open_stream_args` override those inferred
            # from CSV Reader Config
            if open_stream_args:
                prepare_read_kwargs["open_stream_args"] = open_stream_args
            read_tasks = reader.prepare_read(**prepare_read_kwargs)
            all_read_tasks.extend(read_tasks)
        return all_read_tasks

    def prepare_read(
        self,
        parallelism: int,
        paths: Union[str, List[str]],
        content_type_provider: Callable[[str], ContentType],
        path_type: PathType = PathType.MANIFEST,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
        columns: Optional[List[str]] = None,
        schema: Optional[pa.Schema] = None,
        csv_reader_config: DelimitedTextReaderConfig = DelimitedTextReaderConfig(),
        partitioning: HivePartitionParser = None,
        open_stream_args: Optional[Dict[str, Any]] = None,
        read_kwargs_provider: Optional[ReadKwargsProvider] = None,
        **s3_client_kwargs,
    ) -> List[ReadTask]:
        pass
    """



from abc import ABC, abstractmethod
from typing import Iterator, List, Any
import pyarrow as pa

from deltacat.experimental.storage.rivulet.metastore.sst import SSTableRow
from deltacat.experimental.storage.rivulet import Schema
from deltacat.experimental.storage.rivulet.serializer import (
    DataSerializer,
    MEMTABLE_DATA,
)
from deltacat.experimental.storage.rivulet.fs.file_provider import FileProvider


class ArrowSerializer(DataSerializer, ABC):
    """
    Utility class which can serialize data by first converting to arrow as intermediate format
    and then using the provided serialization function
    """

    def __init__(self, file_provider: FileProvider, schema: Schema):
        self.schema = schema
        self.file_provider = file_provider
        self.arrow_schema = self.schema.to_pyarrow()

    @abstractmethod
    def serialize(self, table: pa.Table) -> List[SSTableRow]:
        """
        Write an Arrow table to the specified output file

        :param table: PyArrow table to write
        :return: Number of row groups in the written file
        """
        pass

    def _to_arrow_table(self, sorted_records: MEMTABLE_DATA) -> pa.Table:
        """
        Convert input records to an Arrow table
        """
        if isinstance(sorted_records, pa.Table):
            return sorted_records
        elif isinstance(sorted_records, (Iterator, List)):
            return pa.Table.from_pylist(sorted_records, schema=self.arrow_schema)
        else:
            raise ValueError(f"Unsupported record type: {type(sorted_records)}")

    def _get_min_max_key(self, table: pa.Table) -> (Any, Any):
        """
        Get min and max values for the merge key from the table
        """
        key_col = table[self.schema.get_merge_key()]
        return key_col[0].as_py(), key_col[len(key_col) - 1].as_py()

    def flush_batch(self, sorted_records: MEMTABLE_DATA) -> List[SSTableRow]:
        """
        Write records to new parquet file as row group
        For now, we will depend on pyarrow to write to parquet

        :param sorted_records: record batch in SORTED ORDER
        :return: metadata for constructing SSTable
        """
        if not sorted_records:
            return []

        table = self._to_arrow_table(sorted_records)
        return self.serialize(table)

    def write(self, sorted_records: MEMTABLE_DATA) -> List[SSTableRow]:
        """
        Write records using the provided serialization function

        :param sorted_records: record batch in SORTED ORDER
        :return: metadata for constructing SSTable
        """
        if not sorted_records:
            return []

        table = self._to_arrow_table(sorted_records)
        return self.serialize(table)



import argparse

from datetime import datetime

import ray

import deltacat
import daft
import pyarrow as pa
import pandas as pd
import polars as pl
import numpy as np

from deltacat import DeltaCatUrl


def print_package_version_info() -> None:
    print(f"DeltaCAT Version: {deltacat.__version__}")
    print(f"Ray Version: {ray.__version__}")
    print(f"Daft Version: {daft.__version__}")
    print(f"NumPy Version: {np.__version__}")
    print(f"PyArrow Version: {pa.__version__}")
    print(f"Polars Version: {pl.__version__}")
    print(f"Pandas Version: {pd.__version__}")


def json_path_to_regex(path: str):
    if not path:
        raise ValueError("Path cannot be empty")
    parts = path.split("/")
    leaf_key = parts.pop()
    regex = r""
    for part in parts:
        if part.strip():  # discard leading and/or redundant separators
            regex += rf'"{part}"\s*:\s*[{{\[].*?'
    regex += rf'"{leaf_key}"\s*:\s*"(?<{leaf_key}>.*?)"'
    return regex


def run(
    source: str,
    dest: str,
) -> None:
    # print package version info
    print_package_version_info()

    # run a synchronous copy from the source to the destination
    deltacat.copy(
        DeltaCatUrl(source),
        DeltaCatUrl(dest),
        # reader arguments to pass to the default reader (polars)
        # for the given text-based datasource, it accepts the same
        # arguments as polars.read_csv except for `source`, `n_threads`
        # `new_columns`, `separator`, `has_header`, `quote_char`, and
        # `infer_schema`.
        reader_args={
            "low_memory": True,  # try to use less memory (++stability, --perf)
            "batch_size": 1024,  # text line count read into a buffer at once
            "use_pyarrow": True,  # use the native pyarrow reader
        },
        # writer arguments to pass to the default writer (polars)
        # for the given parquet-based datasink, it generally accepts the same
        # arguments as polars.DataFrame.write_{dest-type} except for `file`
        writer_args={
            "compression": "lz4",  # faster compression & decompression
            # "compression": "zstd",  # better compression ratio
            # "compression": "snappy",  # compatible w/ older Parquet readers
        },
        # Transforms to run against the default polars dataframe read.
        # By default, each transform takes a polars dataframe `df` as input
        # and produces a polars dataframe as output. All transforms listed
        # are run in order (i.e., the dataframe output from transform[0]
        # is the dataframe input to transform[1]).
        #
        # See:
        # https://docs.pola.rs/api/python/stable/reference/dataframe/index.html
        # https://docs.pola.rs/api/python/stable/reference/expressions/index.html
        transforms=[
            lambda df, src: df.rename(
                {"text": "utf8_body"},
            ),
            lambda df, src: df.with_columns(
                pl.col("utf8_body").hash().alias("utf8_body_hash"),
                pl.lit(datetime.utcnow()).dt.datetime().alias("processing_time"),
                pl.lit(src.url_path).alias("source_file_path"),
            ),
        ],
    )


if __name__ == "__main__":
    """
    Example 1: Run this script locally using Ray:
    $ python indexer.py \
    $   --source 'text+s3://openalex-mag-format/data_dump_v1/2022-07-08/nlp/PaperAbstractsInvertedIndex.txt_part31' \
    $   --dest 'parquet+s3://deltacat-example-output/openalex/PaperAbstractsInvertedIndex.part31.parquet'

    Example 2: Submit this script as a local Ray job using a local job client:
    >>> from deltacat import local_job_client
    >>> client = local_job_client()
    >>> # read the source file as line-delimited text
    >>> src = "text+s3://openalex-mag-format/data_dump_v1/2022-07-08/nlp/PaperAbstractsInvertedIndex.txt_part31"
    >>> # write to the destination file using the default DeltaCAT Parquet writer (i.e., polars.DataFrame.write_parquet)
    >>> dst = "parquet+s3://deltacat-example-output/openalex/PaperAbstractsInvertedIndex.part31.parquet"
    >>> try:
    >>>   job_run_result = client.run_job(
    >>>       # Entrypoint shell command to run the indexer job
    >>>       entrypoint=f"python indexer.py --source '{src}' --dest '{dst}'",
    >>>       # Path to the local directory that contains the indexer.py file
    >>>       runtime_env={"working_dir": "./deltacat/examples/indexer.py"},
    >>>   )
    >>>   print(f"Job ID {job_run_result.job_id} terminal state: {job_run_result.job_status}")
    >>>   print(f"Job ID {job_run_result.job_id} logs: ")
    >>>   print(job_run_result.job_logs)
    >>> except RuntimeError as e:
    >>>     print(f"Job Run Failed: {e}")
    >>> except TimeoutError as e:
    >>>     print(f"Job Run Timed Out: {e}")

    Example 3: Submit this script as a remote Ray job using a remote job client:
    >>> from deltacat import job_client
    >>> # use `deltacat.yaml` from the current working directory as the ray cluster launcher config file
    >>> # automatically launches the cluster if it doesn't exist or has died
    >>> # automatically forwards the ray cluster's dashboard for viewing in a web browser @ http://localhost:8265
    >>> client = job_client()
    >>> # ... follow the same steps as above to submit a synchronous indexer job ...
    >>>
    >>> # OR use an explicit cluster launcher config file path
    >>> client = job_client("/Users/pdames/workspace/deltacat.yaml")
    >>> # ... follow the same steps as above to submit a synchronous indexer job ...
    """
    script_args = [
        (
            [
                "--source",
            ],
            {
                "help": "Source DeltaCAT URL to index.",
                "type": str,
            },
        ),
        (
            [
                "--dest",
            ],
            {
                "help": "Destination DeltaCAT URL to index.",
                "type": str,
            },
        ),
    ]
    # parse CLI input arguments
    parser = argparse.ArgumentParser()
    for args, kwargs in script_args:
        parser.add_argument(*args, **kwargs)
    args = parser.parse_args()
    print(f"Command Line Arguments: {args}")

    # initialize deltacat
    deltacat.init()

    # run the example using the parsed arguments
    run(**vars(args))



from deltacat.catalog.delegate import (
    alter_namespace,
    alter_table,
    create_namespace,
    create_table,
    default_namespace,
    drop_namespace,
    drop_table,
    get_namespace,
    get_table,
    list_namespaces,
    list_tables,
    namespace_exists,
    read_table,
    refresh_table,
    rename_table,
    table_exists,
    truncate_table,
    write_to_table,
)
from deltacat.catalog.model.catalog import (  # noqa: F401
    all_catalogs,
    init,
    init_local,
    is_initialized,
    clear_catalogs,
    get_catalog,
    pop_catalog,
    put_catalog,
    raise_if_not_initialized,
    Catalog,
)
from deltacat.catalog.model.properties import (  # noqa: F401
    CatalogProperties,
    get_catalog_properties,
)
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.catalog.main import impl as dcat

__all__ = [
    "alter_namespace",
    "alter_table",
    "create_namespace",
    "create_table",
    "default_namespace",
    "drop_namespace",
    "drop_table",
    "get_namespace",
    "get_table",
    "list_namespaces",
    "list_tables",
    "namespace_exists",
    "read_table",
    "refresh_table",
    "rename_table",
    "table_exists",
    "truncate_table",
    "write_to_table",
    "all_catalogs",
    "init",
    "init_local",
    "is_initialized",
    "clear_catalogs",
    "get_catalog",
    "get_catalog_properties",
    "pop_catalog",
    "put_catalog",
    "raise_if_not_initialized",
    "dcat",
    "Catalog",
    "CatalogProperties",
    "TableDefinition",
]



import tempfile
import shutil
import uuid
import deltacat
import pytest
from deltacat import Field, Schema
from pyiceberg.catalog import CatalogType

import pyarrow as pa

from deltacat.experimental.catalog.iceberg import IcebergCatalogConfig


@pytest.fixture
def schema_a():
    return Schema.of(
        [
            Field.of(
                field=pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )


class TestIcebergCatalogInitialization:
    temp_dir = None

    @classmethod
    def setup_class(cls):
        cls.temp_dir = tempfile.mkdtemp()

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.temp_dir)

    def test_iceberg_catalog_and_table_create(self, schema_a):

        # Register a random catalog name to avoid concurrent test conflicts
        catalog_name = str(uuid.uuid4())

        config = IcebergCatalogConfig(
            type=CatalogType.SQL,
            properties={
                "warehouse": self.temp_dir,
                "uri": f"sqlite:////{self.temp_dir}/sql-catalog.db",
            },
        )

        # Initialize with the PyIceberg catalog
        catalog = deltacat.IcebergCatalog.from_config(config)
        deltacat.init(
            {catalog_name: catalog},
            force=True,
        )

        table_def = deltacat.create_table(
            "test_table", catalog=catalog_name, schema=schema_a
        )

        # Fetch table we just created
        fetched_table_def = deltacat.get_table("test_table", catalog=catalog_name)
        assert table_def.table_version == fetched_table_def.table_version

        # For now, just check that we created a table version with an equivalent schema
        assert table_def.table_version.schema.equivalent_to(schema_a)

        # Sanity check that list namespaces works
        namespaces = deltacat.list_namespaces(catalog=catalog_name).all_items()
        assert table_def.table.namespace in [n.namespace for n in namespaces]



from __future__ import annotations
from typing import Any, Dict

from attr import dataclass
from pyiceberg.catalog import CatalogType


@dataclass
class IcebergCatalogConfig:
    """
    Configuration properties for Iceberg catalog implementation.

    This class holds the PyIceberg Catalog instance needed for interaction with
    Iceberg tables and metadata.

    This configuration is passed through to PyIceberg by invoking load_catalog.
    The Properties provided must match properties accepted by PyIceberg for each catalog type
    See: :func:`deltacat.experimental.catalog.iceberg.initialize`

    Attributes:
        type: The PyIceberg Catalog instance
        properties: Dict of properties passed to pyiceberg load_catalog
    """

    type: CatalogType
    properties: Dict[str, Any]



import tempfile
import os
from typing import Any, Callable, Dict, List, Optional, Set
import pytest
import pyarrow as pa
import ray

from deltacat.io.file_object_store import FileObjectStore
from pytest_benchmark.fixture import BenchmarkFixture

from deltacat.tests.compute.test_util_constant import (
    DEFAULT_NUM_WORKERS,
    DEFAULT_WORKER_INSTANCE_CPUS,
)
from deltacat.tests.compute.test_util_common import (
    get_rci_from_partition,
    read_audit_file,
    PartitionKey,
    get_compacted_delta_locator_from_partition,
)
from deltacat.tests.compute.test_util_common import (
    create_src_w_deltas_destination_rebase_w_deltas_strategy_main,
)

from deltacat.compute.compactor.model.compactor_version import CompactorVersion
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from deltacat.tests.compute.compact_partition_rebase_test_cases import (
    REBASE_TEST_CASES,
)
from deltacat.types.media import StorageType, ContentType
from deltacat.storage import (
    DeltaLocator,
    Partition,
    metastore,
)
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.compute.compactor import (
    RoundCompletionInfo,
)
from deltacat.utils.placement import (
    PlacementGroupManager,
)


"""
MODULE scoped fixtures
"""


@pytest.fixture(autouse=True, scope="module")
def setup_ray_cluster():
    ray.init(local_mode=True, ignore_reinit_error=True)
    yield
    ray.shutdown()


"""
FUNCTION scoped fixtures
"""


@pytest.fixture(autouse=True, scope="function")
def enable_bucketing_spec_validation(monkeypatch):
    """
    Enable the bucketing spec validation for all tests.
    This will help catch hash bucket drift in testing.
    """
    import deltacat.compute.compactor_v2.steps.merge

    monkeypatch.setattr(
        deltacat.compute.compactor_v2.steps.merge,
        "BUCKETING_SPEC_COMPLIANCE_PROFILE",
        "ASSERT",
    )


@pytest.mark.parametrize(
    [
        "test_name",
        "primary_keys",
        "sort_keys",
        "partition_keys_param",
        "partition_values_param",
        "input_deltas_param",
        "input_deltas_delta_type",
        "expected_terminal_compact_partition_result",
        "expected_terminal_exception",
        "expected_terminal_exception_message",
        "create_placement_group_param",
        "records_per_compacted_file_param",
        "hash_bucket_count_param",
        "read_kwargs_provider_param",
        "drop_duplicates_param",
        "skip_enabled_compact_partition_drivers",
        "assert_compaction_audit",
        "rebase_expected_compact_partition_result",
        "compact_partition_func",
        "compactor_version",
    ],
    [
        (
            test_name,
            primary_keys,
            sort_keys,
            partition_keys_param,
            partition_values_param,
            input_deltas,
            input_deltas_delta_type,
            expected_terminal_compact_partition_result,
            expected_terminal_exception,
            expected_terminal_exception_message,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            drop_duplicates_param,
            read_kwargs_provider,
            skip_enabled_compact_partition_drivers,
            assert_compaction_audit,
            rebase_expected_compact_partition_result,
            compact_partition_func,
            compactor_version,
        )
        for test_name, (
            primary_keys,
            sort_keys,
            partition_keys_param,
            partition_values_param,
            input_deltas,
            input_deltas_delta_type,
            expected_terminal_compact_partition_result,
            expected_terminal_exception,
            expected_terminal_exception_message,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            drop_duplicates_param,
            read_kwargs_provider,
            skip_enabled_compact_partition_drivers,
            assert_compaction_audit,
            rebase_expected_compact_partition_result,
            compact_partition_func,
            compactor_version,
        ) in REBASE_TEST_CASES.items()
    ],
    ids=[test_name for test_name in REBASE_TEST_CASES],
)
def test_compact_partition_rebase_same_source_and_destination_main(
    mocker,
    main_deltacat_storage_kwargs: Dict[str, Any],
    test_name: str,
    primary_keys: Set[str],
    sort_keys: List[Optional[Any]],
    partition_keys_param: Optional[List[PartitionKey]],
    partition_values_param: List[Optional[str]],
    input_deltas_param: List[pa.Array],
    input_deltas_delta_type: str,
    expected_terminal_compact_partition_result: pa.Table,
    expected_terminal_exception: BaseException,
    expected_terminal_exception_message: Optional[str],
    create_placement_group_param: bool,
    records_per_compacted_file_param: int,
    hash_bucket_count_param: int,
    drop_duplicates_param: bool,
    read_kwargs_provider_param: Any,
    rebase_expected_compact_partition_result: pa.Table,
    skip_enabled_compact_partition_drivers: List[CompactorVersion],
    assert_compaction_audit: Optional[Callable],
    compactor_version: Optional[CompactorVersion],
    compact_partition_func: Callable,
    benchmark: BenchmarkFixture,
):
    ds_mock_kwargs = main_deltacat_storage_kwargs
    """
    This test tests the scenario where source partition locator == destination partition locator,
    but rebase source partition locator is different.
    This scenario could occur when hash bucket count changes.

    This version uses the main metastore implementation instead of local storage.
    """
    partition_keys = partition_keys_param
    (
        source_table_stream,
        _,
        rebased_table_stream,
    ) = create_src_w_deltas_destination_rebase_w_deltas_strategy_main(
        sort_keys,
        partition_keys,
        input_deltas_param,
        input_deltas_delta_type,
        partition_values_param,
        ds_mock_kwargs,
    )

    # Convert partition values for partition lookup (same as in the helper function)
    converted_partition_values_for_lookup = partition_values_param
    if partition_values_param and partition_keys:
        converted_partition_values_for_lookup = []
        for i, (value, pk) in enumerate(zip(partition_values_param, partition_keys)):
            if pk.key_type.value == "int":  # Use .value to get string representation
                converted_partition_values_for_lookup.append(int(value))
            else:
                converted_partition_values_for_lookup.append(value)

    source_partition: Partition = metastore.get_partition(
        source_table_stream.locator,
        converted_partition_values_for_lookup,
        **ds_mock_kwargs,
    )
    rebased_partition: Partition = metastore.get_partition(
        rebased_table_stream.locator,
        converted_partition_values_for_lookup,
        **ds_mock_kwargs,
    )
    all_column_names = metastore.get_table_version_column_names(
        rebased_table_stream.locator.table_locator.namespace,
        rebased_table_stream.locator.table_locator.table_name,
        rebased_table_stream.locator.table_version_locator.table_version,
        **ds_mock_kwargs,
    )
    num_workers, worker_instance_cpu = DEFAULT_NUM_WORKERS, DEFAULT_WORKER_INSTANCE_CPUS
    total_cpus = num_workers * worker_instance_cpu
    pgm = None
    create_placement_group_param = False
    if create_placement_group_param:
        pgm = PlacementGroupManager(
            1, total_cpus, worker_instance_cpu, memory_per_bundle=4000000
        ).pgs[0]
    last_stream_position_to_compact = source_partition.stream_position
    with tempfile.TemporaryDirectory() as test_dir:
        compact_partition_params = CompactPartitionParams.of(
            {
                "catalog": ds_mock_kwargs.get("inner"),
                "compacted_file_content_type": ContentType.PARQUET,
                "dd_max_parallelism_ratio": 1.0,
                "deltacat_storage": metastore,
                "deltacat_storage_kwargs": ds_mock_kwargs,
                "destination_partition_locator": rebased_partition.locator,
                "hash_bucket_count": hash_bucket_count_param,
                "last_stream_position_to_compact": last_stream_position_to_compact,
                "list_deltas_kwargs": {
                    **ds_mock_kwargs,
                    **{"equivalent_table_types": []},
                },
                "object_store": FileObjectStore(test_dir),
                "pg_config": pgm,
                "primary_keys": primary_keys,
                "all_column_names": all_column_names,
                "read_kwargs_provider": read_kwargs_provider_param,
                "rebase_source_partition_locator": source_partition.locator,
                "rebase_source_partition_high_watermark": rebased_partition.stream_position,
                "records_per_compacted_file": records_per_compacted_file_param,
                "source_partition_locator": rebased_partition.locator,
                "sort_keys": sort_keys if sort_keys else None,
                "drop_duplicates": drop_duplicates_param,
            }
        )

        from deltacat.compute.compactor_v2.model.evaluate_compaction_result import (
            ExecutionCompactionResult,
        )

        execute_compaction_result_spy = mocker.spy(
            ExecutionCompactionResult, "__init__"
        )
        object_store_put_many_spy = mocker.spy(FileObjectStore, "put_many")

        # execute
        benchmark(compact_partition_func, compact_partition_params)

        # Get RoundCompletionInfo from the compacted partition
        round_completion_info: RoundCompletionInfo = get_rci_from_partition(
            rebased_partition.locator, metastore, catalog=ds_mock_kwargs.get("inner")
        )

        # assert if RCI covers all files
        if compactor_version != CompactorVersion.V1.value:
            previous_end = None
            for start, end in round_completion_info.hb_index_to_entry_range.values():
                assert (previous_end is None and start == 0) or start == previous_end
                previous_end = end
            assert (
                previous_end
                == round_completion_info.compacted_pyarrow_write_result.files
            )

        # Get catalog root for audit file resolution
        catalog = ds_mock_kwargs.get("inner")
        catalog_root = catalog.root

        compaction_audit_obj: Dict[str, Any] = read_audit_file(
            round_completion_info.compaction_audit_url, catalog_root
        )
        compaction_audit: CompactionSessionAuditInfo = CompactionSessionAuditInfo(
            **compaction_audit_obj
        )

        # Assert not in-place compacted
        assert (
            execute_compaction_result_spy.call_args.args[-1] is False
        ), "Table version erroneously marked as in-place compacted!"
        compacted_delta_locator: DeltaLocator = (
            get_compacted_delta_locator_from_partition(
                rebased_partition.locator,
                metastore,
                catalog=ds_mock_kwargs.get("inner"),
            )
        )
        assert (
            compacted_delta_locator.stream_position == last_stream_position_to_compact
        ), "Compacted delta locator must be equal to last stream position"
        tables = metastore.download_delta(
            compacted_delta_locator, storage_type=StorageType.LOCAL, **ds_mock_kwargs
        )
        actual_rebase_compacted_table = pa.concat_tables(tables)
        # if no primary key is specified then sort by sort_key for consistent assertion
        sorting_cols: List[Any] = []
        if primary_keys:
            sorting_cols.extend([(val, "ascending") for val in primary_keys])
        if sort_keys:
            sorting_cols.extend([pa_key for key in sort_keys for pa_key in key.arrow])

        rebase_expected_compact_partition_result = (
            rebase_expected_compact_partition_result.combine_chunks().sort_by(
                sorting_cols
            )
        )
        actual_rebase_compacted_table = (
            actual_rebase_compacted_table.combine_chunks().sort_by(sorting_cols)
        )
        assert actual_rebase_compacted_table.equals(
            rebase_expected_compact_partition_result
        ), f"{actual_rebase_compacted_table} does not match {rebase_expected_compact_partition_result}"

        if assert_compaction_audit is not None:
            if not assert_compaction_audit(compactor_version, compaction_audit):
                assert False, "Compaction audit assertion failed"
        # We do not expect object store to be cleaned up when there's only one round
        if object_store_put_many_spy.call_count:
            assert os.listdir(test_dir) != []



from deltacat.storage.model.expression.expression import (
    Expression,
    UnaryExpression,
    BinaryExpression,
    BooleanExpression,
    Reference,
    Literal,
    Equal,
    NotEqual,
    GreaterThan,
    LessThan,
    GreaterThanEqual,
    LessThanEqual,
    And,
    Or,
    Not,
    In,
    Between,
    Like,
    IsNull,
)

from deltacat.storage.model.expression.visitor import ExpressionVisitor, DisplayVisitor

__all__ = [
    "Expression",
    "UnaryExpression",
    "BinaryExpression",
    "BooleanExpression",
    "Reference",
    "Literal",
    "Equal",
    "NotEqual",
    "GreaterThan",
    "LessThan",
    "GreaterThanEqual",
    "LessThanEqual",
    "And",
    "Or",
    "Not",
    "In",
    "Between",
    "Like",
    "IsNull",
    "ExpressionVisitor",
    "DisplayVisitor",
]



from unittest.mock import Mock
from deltacat.compute.compactor.utils.round_completion_reader import (
    read_round_completion_info,
)
from deltacat.tests.compute.test_util_common import get_test_partition_locator
from deltacat.compute.compactor import RoundCompletionInfo
from deltacat.compute.compactor.model.pyarrow_write_result import PyArrowWriteResult
from deltacat.storage.model.partition import Partition


class TestRoundCompletionInfoInPartition:
    def test_read_round_completion_info_from_partition_with_matching_source(self):
        """
        Test reading RoundCompletionInfo from partition metafile with matching source partition locator.
        """
        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        # Create a test RoundCompletionInfo with prev_source_partition_locator
        pawr = PyArrowWriteResult.of(
            file_count=1, pyarrow_bytes=1000, file_bytes=1000, record_count=100
        )

        expected_rci = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator=None,
            compacted_pyarrow_write_result=pawr,
            sort_keys_bit_width=12,
            prev_source_partition_locator=source_locator,
        )

        # Create a partition with RoundCompletionInfo
        partition = Partition.of(
            locator=destination_locator,
            content_types=None,
            compaction_round_completion_info=expected_rci,
        )

        # Mock the storage
        mock_storage = Mock()

        # Test reading with partition provided (no storage call needed)
        rci = read_round_completion_info(
            source_partition_locator=source_locator,
            destination_partition_locator=destination_locator,
            deltacat_storage=mock_storage,
            deltacat_storage_kwargs={},
            destination_partition=partition,
        )

        assert rci is not None
        assert rci == expected_rci
        assert rci.high_watermark == 122
        assert rci.sort_keys_bit_width == 12
        assert (
            rci.prev_source_partition_locator.partition_id
            == source_locator.partition_id
        )

        # Verify storage was not called since partition was provided
        mock_storage.get_partition.assert_not_called()

    def test_read_round_completion_info_from_partition_with_mismatched_source(self):
        """
        Test reading RoundCompletionInfo from partition metafile with mismatched source partition locator.
        Should return None and log a warning.
        """
        source_locator = get_test_partition_locator("source")
        different_source_locator = get_test_partition_locator("different_source")
        destination_locator = get_test_partition_locator("destination")

        # Create a test RoundCompletionInfo with different prev_source_partition_locator
        pawr = PyArrowWriteResult.of(
            file_count=1, pyarrow_bytes=1000, file_bytes=1000, record_count=100
        )

        expected_rci = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator=None,
            compacted_pyarrow_write_result=pawr,
            sort_keys_bit_width=12,
            prev_source_partition_locator=different_source_locator,  # Different from source_locator
        )

        # Create a partition with RoundCompletionInfo
        partition = Partition.of(
            locator=destination_locator,
            content_types=None,
            compaction_round_completion_info=expected_rci,
        )

        # Mock the storage
        mock_storage = Mock()

        # Test reading with mismatched source locator
        rci = read_round_completion_info(
            source_partition_locator=source_locator,  # Different from the one in RoundCompletionInfo
            destination_partition_locator=destination_locator,
            deltacat_storage=mock_storage,
            deltacat_storage_kwargs={},
            destination_partition=partition,
        )

        # Should return None due to mismatch
        assert rci is None

        # Verify storage was not called since partition was provided
        mock_storage.get_partition.assert_not_called()

    def test_read_round_completion_info_from_storage_when_partition_not_provided(self):
        """
        Test reading RoundCompletionInfo from storage when partition is not provided.
        """
        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        # Create a test RoundCompletionInfo
        pawr = PyArrowWriteResult.of(
            file_count=1, pyarrow_bytes=1000, file_bytes=1000, record_count=100
        )

        expected_rci = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator=None,
            compacted_pyarrow_write_result=pawr,
            sort_keys_bit_width=12,
            prev_source_partition_locator=source_locator,
        )

        # Create a partition with RoundCompletionInfo
        partition = Partition.of(
            locator=destination_locator,
            content_types=None,
            compaction_round_completion_info=expected_rci,
        )

        # Mock the storage to return the partition
        mock_storage = Mock()
        mock_storage.get_partition.return_value = partition

        # Test reading without partition provided (storage call needed)
        rci = read_round_completion_info(
            source_partition_locator=source_locator,
            destination_partition_locator=destination_locator,
            deltacat_storage=mock_storage,
            deltacat_storage_kwargs={"test_arg": "test_value"},
        )

        assert rci is not None
        assert rci == expected_rci
        assert rci.high_watermark == 122

        # Verify storage was called with correct parameters
        mock_storage.get_partition.assert_called_once_with(
            destination_locator.stream_locator,
            destination_locator.partition_values,
            test_arg="test_value",
        )

    def test_read_round_completion_info_when_partition_not_found(self):
        """
        Test reading RoundCompletionInfo when partition is not found in storage.
        """
        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        # Mock the storage to return None (partition not found)
        mock_storage = Mock()
        mock_storage.get_partition.return_value = None

        # Test reading when partition not found
        rci = read_round_completion_info(
            source_partition_locator=source_locator,
            destination_partition_locator=destination_locator,
            deltacat_storage=mock_storage,
            deltacat_storage_kwargs={},
        )

        # Should return None when partition not found
        assert rci is None

        # Verify storage was called
        mock_storage.get_partition.assert_called_once()

    def test_read_round_completion_info_when_no_completion_info_in_partition(self):
        """
        Test reading RoundCompletionInfo when partition exists but has no completion info.
        """
        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        # Create a partition without RoundCompletionInfo
        partition = Partition.of(
            locator=destination_locator,
            content_types=None,
            compaction_round_completion_info=None,
        )

        # Mock the storage to return the partition
        mock_storage = Mock()
        mock_storage.get_partition.return_value = partition

        # Test reading when no completion info in partition
        rci = read_round_completion_info(
            source_partition_locator=source_locator,
            destination_partition_locator=destination_locator,
            deltacat_storage=mock_storage,
            deltacat_storage_kwargs={},
        )

        # Should return None when no completion info
        assert rci is None

    def test_read_with_missing_prev_source_partition_locator_returns_none(self):
        """
        Test that reading with missing prev_source_partition_locator returns None.
        """
        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        # Create RoundCompletionInfo without prev_source_partition_locator
        pawr = PyArrowWriteResult.of(
            file_count=1, pyarrow_bytes=1000, file_bytes=1000, record_count=100
        )

        rcf = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator=None,
            compacted_pyarrow_write_result=pawr,
            sort_keys_bit_width=12,
            prev_source_partition_locator=None,  # Missing
        )

        # Create a partition with RoundCompletionInfo
        partition = Partition.of(
            locator=destination_locator,
            content_types=None,
            compaction_round_completion_info=rcf,
        )

        # Mock the storage
        mock_storage = Mock()

        # Test reading should return None due to missing prev_source_partition_locator
        result = read_round_completion_info(
            source_partition_locator=source_locator,
            destination_partition_locator=destination_locator,
            deltacat_storage=mock_storage,
            deltacat_storage_kwargs={},
            destination_partition=partition,
        )

        # Should return None when prev_source_partition_locator is missing or mismatched
        assert result is None
